from __future__ import annotations
from bs_monitoring.alert_services.base import (
    AlertService,
    alert,
    create_alert_service,
    register_alert_service,
    AlertServiceConfig,
)

from bs_monitoring.alert_services.opsgenie import OpsgenieAlertService
from bs_monitoring.alert_services.discord import DiscordAlertService

__all__ = [
    "AlertService",
    "AlertServiceConfig",
    "alert",
    "create_alert_service",
    "register_alert_service",
]



from typing import Any
from bs_monitoring.alert_services import alert, AlertService
from bs_monitoring.common.utils import MonitoringServiceError
import pytz
from datetime import datetime, timedelta

from bs_monitoring.monitors import Monitor, register_monitor


class DataQuantityError(MonitoringServiceError):
    def __init__(self, message: str) -> None:
        """Exception raised when data quantity is 0.

        Args:
            message (str): The error message.
        """
        super().__init__(message)


@register_monitor
class DataQuantityMonitor(Monitor):
    def __init__(
        self,
        alert_service: AlertService,
        config: None = None,
        db_name: str | None = None,
    ) -> None:
        """Monitor to check the quantity of data.

        Args:
            alert_service (AlertService): The alert service to use.
            config (None, optional): No purpose, required because of the factory method. Defaults to None.
        """
        super().__init__(alert_service, db_name)

    @alert(message="Data quantity error")
    async def process(self, data: dict[str, Any]) -> None:
        """Method to process the data, sends an alert if the data date is a weekday and the data quantity is 0.

        Args:
            data (Dict[str, Any]): The data to process.

        Raises:
            DataQuantityError: Raised if the data quantity is 0 and is weekday.
        """
        start_date = datetime.now(pytz.utc) - timedelta(days=1)
        if start_date.weekday() >= 5:
            return

        for k, v in data.items():
            if len(v) == 0:
                raise DataQuantityError(f"Data quantity for {k} is 0")



from dataclasses import dataclass
import sqlite3

from bs_monitoring.common.utils import ConfigField
from bs_monitoring.databases.base import Database, register_database


@dataclass
class SqliteDatabaseConfig:
    """Configuration for the SQLite database.

    Args:
        path (str): The path to the SQLite database.
    """

    path: str


@register_database
class SqliteDatabase(Database):
    path = ConfigField(str)

    def __init__(self, config: SqliteDatabaseConfig) -> None:
        """SQLite database implementation.

        Args:
            config (SqliteDatabaseConfig): The configuration for the database.
        """
        super().__init__(config)
        self.conn = None

    def connect(self) -> None:
        self.conn = sqlite3.connect(self.path)

    def close(self) -> None:
        if self.conn is not None:
            self.conn.close()

    def execute(self, query: str, params=None):
        cursor = self.conn.cursor()

        if params is not None:
            cursor.execute(query, params)
        else:
            cursor.execute(query)

        return cursor

    def commit(self) -> None:
        self.conn.commit()

    @property
    def autocommit(self) -> bool:
        return self.conn.autocommit

    @autocommit.setter
    def autocommit(self, value: bool) -> None:
        self.conn.autocommit = value



from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import Any

from bs_monitoring.common.utils import register_config


@dataclass
class DatabaseConfig:
    type: str
    name: str
    config: None = None


class Database(ABC):
    def __init__(self, config: Any):
        for k, v in asdict(config).items():
            setattr(self, k, v)

    @abstractmethod
    def connect(self) -> None:
        """Method to connect to the database."""
        pass

    @abstractmethod
    def close(self) -> None:
        """Method to close the connection to the database."""
        pass

    @abstractmethod
    def execute(self, query: str) -> None:
        """Method to execute a query on the database.

        Args:
            query (str): The query to execute.
        """
        pass

    @abstractmethod
    def commit(self) -> None:
        """Method to commit the transaction."""
        pass

    @property
    def autocommit(self) -> bool:
        """Property to get the autocommit status of the connection."""
        pass

    @autocommit.setter
    def autocommit(self, value: bool) -> None:
        """Property to set the autocommit status of the connection.

        Args:
            value (bool): The value to set.
        """
        pass


def register_database(cls):
    assert issubclass(
        cls, Database
    ), f"Class {cls.__name__} must be a subclass of Database"

    name = (
        cls.__name__
        if not cls.__name__.endswith("Database")
        else cls.__name__.removesuffix("Database")
    )
    DATABASES[name] = cls
    register_config(cls, DatabaseConfig)
    return cls


DATABASES: dict[str, Database] = {}


class DatabaseContext:
    _instance = None
    _db_connections: dict[str, Database] = {}

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(DatabaseContext, cls).__new__(cls)
        return cls._instance

    @classmethod
    def initialize(cls, name, db_type, config):
        conn: Database = DATABASES[db_type](config)
        conn.connect()
        cls._db_connections[name] = conn

    @classmethod
    def get_connection(cls, name):
        if name not in cls._db_connections:
            raise RuntimeError(f"Database connection {name} not initialized")
        return cls._db_connections[name]

    @classmethod
    def close_all(cls):
        for c in cls._db_connections.values():
            c.close()
        cls._db_connections.clear()


def initialize_databases(configs: list[DatabaseConfig]) -> None:
    for config in configs:
        DatabaseContext.initialize(config.name, config.type, config.config)



from bs_monitoring.monitors.base import (
    create_monitors,
    register_monitor,
    Monitor,
    MonitorConfig,
)

from bs_monitoring.monitors.data_quantity import DataQuantityMonitor
from bs_monitoring.monitors.data_scheme import DataSchemeMonitor


__all__ = [
    "create_monitors",
    "register_monitor",
    "Monitor",
    "MonitorConfig",
]



from __future__ import annotations
from bs_monitoring.databases.base import (
    Database,
    register_database,
    DatabaseContext,
    DatabaseConfig,
    initialize_databases,
)

from bs_monitoring.databases.sqlite import SqliteDatabase
from bs_monitoring.databases.postgres import PostgresDatabase


__all__ = [
    "Database",
    "register_database",
    "initialize_databases",
    "DatabaseContext",
    "DatabaseConfig",
]



from __future__ import annotations
from bs_monitoring.data_sources.base import (
    DataSource,
    DataSourceConfig,
    register_data_source,
    create_data_source,
)
from bs_monitoring.data_sources.elastic import ElasticDataSource


__all__ = [
    "DataSource",
    "DataSourceConfig",
    "create_data_source",
    "register_data_source",
]



from bs_monitoring.monitors import create_monitors, Monitor
from bs_monitoring.alert_services import create_alert_service
from bs_monitoring.data_sources import create_data_source, DataSource
from bs_monitoring.databases import initialize_databases
from bs_monitoring.common.configs.base import RootConfig
import asyncio
import signal
from contextlib import AsyncExitStack


class Pipeline:
    def __init__(self, source: DataSource, monitors: list[Monitor], interval: int):
        self.source = source
        self.monitors = monitors
        self._exit_stack = AsyncExitStack()
        self._shutdown_event = asyncio.Event()
        self._interval = interval

    async def shutdown(self, signal=None):
        """Cleanup tasks tied to the service's shutdown."""

        self._shutdown_event.set()

        await self.source.close()

        for monitor in self.monitors:
            if hasattr(monitor, "close"):
                await monitor.close()

        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        for task in tasks:
            task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)

    @staticmethod
    def construct(config: RootConfig):
        initialize_databases(config.Databases)

        alert_service = create_alert_service(config.AlertService)
        source = create_data_source(config.DataSource, alert_service)
        monitors = create_monitors(config.Monitors, alert_service)

        return Pipeline(source, monitors, config.Interval)

    async def _run(self):
        """Run the pipeline continuously using asyncio."""
        try:
            async for data in self.source.produce_continuous(self._interval):
                if self._shutdown_event.is_set():
                    break
                await asyncio.gather(
                    *[monitor.process(data) for monitor in self.monitors]
                )
        except asyncio.CancelledError:
            self._shutdown_event.set()
            return

    def run(self):
        """Entry point to run the pipeline in the event loop."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        # Handle signals
        signals = (signal.SIGTERM, signal.SIGINT)
        for s in signals:
            loop.add_signal_handler(
                s, lambda s=s: asyncio.create_task(self.shutdown(signal=s))
            )

        try:
            loop.run_until_complete(self._run())
        except KeyboardInterrupt:
            loop.run_until_complete(self.shutdown())
        finally:
            tasks = [t for t in asyncio.all_tasks(loop) if not t.done()]
            if tasks:
                loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
            loop.close()
